// Abstract Syntax tree for the MUESLI interpreter
package muesli

import (
	"fmt"
	"strings"
)

/*
type AstKind int

const (
	AstKindProgram = AstKind(iota)
	AstKindStatements
	AstKindStatement
	AstKindSet
	AstKindGet
	AstKindTarget
	AstKindCommand
	AstKindArguments
	AstKindArgument
	AstKindExpression
	AstKindBlock
	AstKindParenthesis
	AstKindList
	AstKindCapture
	AstKindWordValue
	AstKindWord
	AstKindType
	AstKindValue
	AstKindEnd
	AstKindError
)
*/

/*
func (BasicAst Ast) String() string {
	switch astkind {
	case AstKindProgram:
		return "AstKindProgram"
	case AstKindStatements:
		return "AstKindStatements"
	case AstKindStatement:
		return "AstKindStatement"
	case AstKindSet:
		return "AstKindSet"
	case AstKindGet:
		return "AstKindGet"
	case AstKindTarget:
		return "AstKindTarget"
	case AstKindCommand:
		return "AstKindCommand"
	case AstKindArguments:
		return "AstKindArguments"
	case AstKindArgument:
		return "AstKindArgument"
	case AstKindExpression:
		return "AstKindExpression"
	case AstKindBlock:
		return "AstKindBlock"
	case AstKindParenthesis:
		return "AstKindParenthesis"
	case AstKindList:
		return "AstKindList"
	case AstKindCapture:
		return "AstKindCapture"
	case AstKindWordValue:
		return "AstKindWordValue"
	case AstKindWord:
		return "AstKindWord"
	case AstKindType:
		return "AstKindType"
	case AstKindValue:
		return "AstKindValue"
	case AstKindEnd:
		return "AstKindEnd"
	case AstKindError:
		return "AstKindError"
	default:
		return "Unknown AstKind"
	}
}
*/

type Ast interface {
	Parent() Ast
	Children() []Ast

	Run(vm *VM, val ListValue) ListValue
	String() string
	Token() Token
	/* AppendChild(child Ast) Ast
	NewChild(kind AstKind, token Token) Ast
	*/

	/* Make(parent Ast, token Token) Ast */
}

/*
func AstAppendChild(parent Ast, child Ast) Ast {
	basicParent := parent.(*BasicAst)
	return basicParent.AppendChild(child)
}

func AstNewChild(parent Ast, kind AstKind, token Token) Ast {
	basicParent := parent.(*BasicAst)
	return basicParent.NewChild(kind, token)
}
*/

/* AST node kind */
type BasicAst struct {
	parent   Ast
	children []Ast
	token    Token
}

var _ Ast = &BasicAst{}

func NewAst(kind AstKind, parent Ast, children []Ast, token Token) Ast {
	ast := kind(parent, children, token)
	return ast
}

/*
func (ast *BasicAst) AppendChild(child Ast) Ast {
	ast.children = append(ast.children, child)
	return child
}

func (ast *BasicAst) NewChild(kind AstKind, token Token) Ast {
	child := NewAst(kind, ast, token)
	ast.AppendChild(child)
	return child
}
*/

func (ast BasicAst) IsError() bool {
	return false
}

func (ast AstError) IsError() bool {
	return true
}

func (ast BasicAst) Token() Token {
	return ast.token
}

func (ast BasicAst) Parent() Ast {
	return ast.parent
}

func (ast BasicAst) Children() []Ast {
	return ast.children
}

func (ast *BasicAst) SetParent(parent Ast) {
	ast.parent = parent
}

func (ast BasicAst) Child(index int) Ast {
	count := ast.CountChildren()
	if index < 0 || index > count {
		return nil
	}
	return ast.children[index]
}

func (ast BasicAst) Run(vm *VM, val ListValue) ListValue {
	panic("Please implement Run for this node.")
	return ListValue{}
}

func (ast *BasicAst) Walk(walker func(node Ast) Ast) Ast {
	if found := walker(ast); found != nil {
		return found
	}
	for _, child := range ast.children {
		if found := child.(*BasicAst).Walk(walker); found != nil {
			return found
		}
	}
	return nil
}

func (ast BasicAst) String() string {
	if ast.Token == nil {
		return fmt.Sprintf("BasicAst %s nil", ast.token.String())
	}
	return fmt.Sprintf("BasicAst %v", ast.token.String())
}

func (ast BasicAst) Display() {
	ast.Walk(func(node Ast) Ast {
		bast := node.(*BasicAst)
		depth := bast.Depth()
		fmt.Printf(strings.Repeat("--", depth))
		if node != nil {
			fmt.Printf("Ast: %s\n", node.String())
		} else {
			fmt.Printf("Ast: nil node\n")
		}
		return nil
	})
}

func (ast BasicAst) Depth() int {
	var depth int = 0
	parent := ast.parent
	for parent != nil {
		depth++
		parent = parent.Parent()
	}
	return depth
}

func (ast BasicAst) CountChildren() int {
	return len(ast.children)
}

func AstIsError(ast Ast) bool {
	_, isError := ast.(*AstError)
	return isError
}

func (ast BasicAst) Errors() []Ast {
	res := make([]Ast, 0)
	ast.Walk(func(node Ast) Ast {
		if node != nil && ast.IsError() {
			res = append(res, node)
		}
		return nil
	})
	return res
}

func EmptyAstArray() []Ast {
	return make([]Ast, 0)
}

type AstProgram struct{ BasicAst }
type AstStatements struct{ BasicAst }
type AstStatement struct{ BasicAst }
type AstSet struct{ BasicAst }
type AstGet struct{ BasicAst }
type AstTarget struct{ BasicAst }
type AstCommand struct{ BasicAst }
type AstArguments struct{ BasicAst }
type AstArgument struct{ BasicAst }
type AstExpression struct{ BasicAst }
type AstBlock struct{ BasicAst }
type AstParenthesis struct{ BasicAst }
type AstList struct{ BasicAst }
type AstCapture struct{ BasicAst }
type AstWordValue struct{ BasicAst }
type AstWord struct{ BasicAst }
type AstType struct{ BasicAst }
type AstValue struct{ BasicAst }
type AstEnd struct{ BasicAst }
type AstError struct{ BasicAst }

func (ast AstProgram) String() string     { return "AstProgram    " }
func (ast AstStatements) String() string  { return "AstStatements " }
func (ast AstStatement) String() string   { return "AstStatement  " }
func (ast AstSet) String() string         { return "AstSet        " }
func (ast AstGet) String() string         { return "AstGet        " }
func (ast AstTarget) String() string      { return "AstTarget     " }
func (ast AstCommand) String() string     { return "AstCommand    " }
func (ast AstArguments) String() string   { return "AstArguments  " }
func (ast AstArgument) String() string    { return "AstArgument   " }
func (ast AstExpression) String() string  { return "AstExpression " }
func (ast AstBlock) String() string       { return "AstBlock      " }
func (ast AstParenthesis) String() string { return "AstParenthesis" }
func (ast AstList) String() string        { return "AstList       " }
func (ast AstCapture) String() string     { return "AstCapture    " }
func (ast AstWordValue) String() string   { return "AstWordValue  " }
func (ast AstWord) String() string        { return "AstWord       " }
func (ast AstType) String() string        { return "AstType       " }
func (ast AstValue) String() string       { return "AstValue      " }
func (ast AstEnd) String() string         { return "AstEnd        " }
func (ast AstError) String() string       { return "AstError      " }

var _ Ast = AstProgram{}

func NewBasicAst(parent Ast, children []Ast, token Token) BasicAst {
	return BasicAst{parent, children, token}
}

/* The kind of an AST is also it's constructor function. This is so deep I'm almost
getting scared. :) */
// The type is defined as a constructor
type AstKind func(parent Ast, children []Ast, token Token) Ast

type AstKind2 interface {
	New(parent Ast, children []Ast, token Token) Ast
}

type AstTypeProgramImp struct{}

func (ap AstTypeProgramImp) New(parent Ast, children []Ast, token Token) Ast {
	return AstProgram{NewBasicAst(parent, children, token)}
}

func NewAstProgram(parent Ast, children []Ast, token Token) Ast {
	return AstProgram{NewBasicAst(parent, children, token)}
}

var AstKindBasic = AstKind(func(parent Ast, children []Ast, token Token) Ast { return NewBasicAst(parent, children, token) })

var AstKindProgram = AstKind(NewAstProgram)
var AstKindProgram2 = AstKind2(AstTypeProgramImp{})

var AstKindStatements = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstStatements{NewBasicAst(parent, children, token)}
})

var AstKindStatement = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstStatement{NewBasicAst(parent, children, token)}
})
var AstKindSet = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstSet{NewBasicAst(parent, children, token)}
})
var AstKindGet = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstGet{NewBasicAst(parent, children, token)}
})
var AstKindTarget = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstTarget{NewBasicAst(parent, children, token)}
})
var AstKindCommand = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstCommand{NewBasicAst(parent, children, token)}
})
var AstKindArguments = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstArguments{NewBasicAst(parent, children, token)}
})
var AstKindArgument = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstArgument{NewBasicAst(parent, children, token)}
})
var AstKindExpression = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstExpression{NewBasicAst(parent, children, token)}
})
var AstKindBlock = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstBlock{NewBasicAst(parent, children, token)}
})
var AstKindParenthesis = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstParenthesis{NewBasicAst(parent, children, token)}
})
var AstKindList = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstList{NewBasicAst(parent, children, token)}
})
var AstKindCapture = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstCapture{NewBasicAst(parent, children, token)}
})
var AstKindWordValue = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstWordValue{NewBasicAst(parent, children, token)}
})
var AstKindWord = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstWord{NewBasicAst(parent, children, token)}
})
var AstKindType = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstType{NewBasicAst(parent, children, token)}
})
var AstKindValue = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstValue{NewBasicAst(parent, children, token)}
})
var AstKindEnd = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstEnd{NewBasicAst(parent, children, token)}
})
var AstKindError = AstKind(func(parent Ast, children []Ast, token Token) Ast {
	return &AstError{NewBasicAst(parent, children, token)}
})
